"""
This module holds tsetmc modules, which are mostly \
inherited from default models.
"""
from dataclasses import dataclass
from datetime import datetime, date, time
from tse_utils.models.enums import Nsc
from tse_utils.models import realtime, instrument


@dataclass
class InstrumentIdentification(instrument.InstrumentIdentification):
    """Identification information of an instrument"""
    market_code: int = None
    market_title: str = None
    sector_code: int = None
    sector_title: str = None
    sub_sector_code: int = None
    sub_sector_title: str = None
    type_id: int = None

    def __init__(self, tsetmc_code: str, tsetmc_raw_data: dict):
        self.market_title = tsetmc_raw_data["cgrValCotTitle"]
        self.market_code = int(tsetmc_raw_data["cComVal"])
        self.sector_code = int(
            tsetmc_raw_data["sector"]["cSecVal"].replace(" ", "")
        )
        self.sector_title = tsetmc_raw_data["sector"]["lSecVal"]
        if "subSector" in tsetmc_raw_data:
            self.sub_sector_code = int(
                tsetmc_raw_data["subSector"]["cSoSecVal"]
            )
            self.sub_sector_title = tsetmc_raw_data["subSector"]["lSoSecVal"]
        self.type_id = int(tsetmc_raw_data["yVal"])
        instrument.InstrumentIdentification.__init__(
            self=self,
            tsetmc_code=tsetmc_code,
            isin=tsetmc_raw_data["instrumentID"],
            ticker=tsetmc_raw_data["lVal18AFC"],
            name_persian=tsetmc_raw_data["lVal30"],
            name_english=tsetmc_raw_data["lVal18"]
        )

    def __str__(self):
        return f"{self.ticker} [{self.isin}]"


@dataclass
class InstrumentSearchItem:
    """An item from the search results in TSETMC"""
    ticker: str = None
    name_persian: str = None
    tsetmc_code: str = None
    is_active: bool = None
    market_title: str = None

    def __init__(self, tsetmc_raw_data: dict):
        self.ticker = tsetmc_raw_data["lVal18AFC"]
        self.name_persian = tsetmc_raw_data["lVal30"]
        self.tsetmc_code = tsetmc_raw_data["insCode"]
        self.market_title = tsetmc_raw_data["flowTitle"]
        self.is_active = tsetmc_raw_data["lastDate"] != 0

    def __str__(self):
        return f"{self.ticker} <{self.tsetmc_code}>"


@dataclass
class ClosingPriceInfo(realtime.TradeCandle):
    """Holds realtime trade data of an instrument"""
    nsc: Nsc = None

    def __init__(self, tsetmc_raw_data: dict):
        self.nsc = Nsc[
            str(
                tsetmc_raw_data["instrumentState"]["cEtaval"]
            ).replace(" ", "")
        ]
        ltd_date = tsetmc_raw_data["finalLastDate"]
        ltd_time = tsetmc_raw_data["hEven"]
        last_trade_datetime = datetime(
            year=ltd_date // 10000, month=ltd_date // 100 % 100, day=ltd_date % 100,
            hour=ltd_time // 10000, minute=ltd_time // 100 % 100, second=ltd_time % 100
        )
        realtime.TradeCandle.__init__(
            self=self,
            previous_price=tsetmc_raw_data["priceYesterday"],
            open_price=tsetmc_raw_data["priceFirst"],
            last_price=tsetmc_raw_data["pDrCotVal"],
            close_price=tsetmc_raw_data["pClosing"],
            max_price=tsetmc_raw_data["priceMax"],
            min_price=tsetmc_raw_data["priceMin"],
            trade_num=tsetmc_raw_data["zTotTran"],
            trade_value=tsetmc_raw_data["qTotCap"],
            trade_volume=tsetmc_raw_data["qTotTran5J"],
            last_trade_datetime=last_trade_datetime
        )


@dataclass
class InstrumentInfo(InstrumentIdentification, realtime.BigQuantityParams):
    """Contains the instrument homepage data"""
    liquid_percentage: float = None
    average_trade_volume_monthly: int = None
    price_range_weekly: realtime.PriceRange = None
    price_range_annual: realtime.PriceRange = None
    price_thresholds: realtime.PriceRange = None

    def __init__(self, tsetmc_code: str, tsetmc_raw_data=dict):
        InstrumentIdentification.__init__(
            self=self,
            tsetmc_code=tsetmc_code,
            tsetmc_raw_data=tsetmc_raw_data
        )
        realtime.BigQuantityParams.__init__(
            self=self,
            base_volume=tsetmc_raw_data["baseVol"],
            total_shares=tsetmc_raw_data["zTitad"]
        )
        self.price_thresholds = realtime.PriceRange(
            max_price=tsetmc_raw_data["staticThreshold"]["psGelStaMax"],
            min_price=tsetmc_raw_data["staticThreshold"]["psGelStaMin"]
        )
        self.price_range_weekly = realtime.PriceRange(
            max_price=tsetmc_raw_data["maxWeek"],
            min_price=tsetmc_raw_data["minWeek"]
        )
        self.price_range_annual = realtime.PriceRange(
            max_price=tsetmc_raw_data["maxYear"],
            min_price=tsetmc_raw_data["minYear"]
        )
        self.average_trade_volume_monthly = tsetmc_raw_data["qTotTran5JAvg"]
        self.liquid_percentage = None \
            if len(tsetmc_raw_data["kAjCapValCpsIdx"]) == 0 \
            else float(tsetmc_raw_data["kAjCapValCpsIdx"])


@dataclass
class ClientType(realtime.ClientType):
    """ClientType data for an instrument"""

    def __init__(self, tsetmc_raw_data):
        realtime.ClientType.__init__(
            self=self,
            legal=realtime.ClientTypeTrade(
                buy=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["buy_CountN"],
                    volume=tsetmc_raw_data["buy_N_Volume"]
                ),
                sell=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["sell_CountN"],
                    volume=tsetmc_raw_data["sell_N_Volume"]
                )
            ),
            natural=realtime.ClientTypeTrade(
                buy=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["buy_CountI"],
                    volume=tsetmc_raw_data["buy_I_Volume"]
                ),
                sell=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["sell_CountI"],
                    volume=tsetmc_raw_data["sell_I_Volume"]
                )
            )
        )


@dataclass
class BestLimitsRow(realtime.OrderBookRow):
    """A single row from an instrument's order book"""

    def __init__(self, tsetmc_raw_data):
        realtime.OrderBookRow.__init__(
            self=self,
            demand=realtime.OrderBookRowSide(
                num=tsetmc_raw_data["zOrdMeDem"],
                volume=tsetmc_raw_data["qTitMeDem"],
                price=tsetmc_raw_data["pMeDem"]
            ),
            supply=realtime.OrderBookRowSide(
                num=tsetmc_raw_data["zOrdMeOf"],
                volume=tsetmc_raw_data["qTitMeOf"],
                price=tsetmc_raw_data["pMeOf"]
            )
        )


@dataclass
class BestLimits:
    """
    BestLimit is the same as OrderBook, contains an instrument's active orders
    """
    rows: list[BestLimitsRow] = None

    def __init__(self, tsetmc_raw_data):
        self.rows = [BestLimitsRow(x) for x in tsetmc_raw_data]


@dataclass
class ClosingPriceDaily(realtime.TradeCandle):
    """Holds historical data of an instrument's trades for a single day"""

    def __init__(self, tsetmc_raw_data):
        ltd_date = tsetmc_raw_data["dEven"]
        ltd_time = tsetmc_raw_data["hEven"]
        last_trade_datetime = datetime(
            year=ltd_date // 10000, month=ltd_date // 100 % 100, day=ltd_date % 100,
            hour=ltd_time // 10000, minute=ltd_time // 100 % 100, second=ltd_time % 100
        )
        realtime.TradeCandle.__init__(
            self=self,
            previous_price=tsetmc_raw_data["priceYesterday"],
            open_price=tsetmc_raw_data["priceFirst"],
            last_price=tsetmc_raw_data["pDrCotVal"],
            close_price=tsetmc_raw_data["pClosing"],
            max_price=tsetmc_raw_data["priceMax"],
            min_price=tsetmc_raw_data["priceMin"],
            trade_num=tsetmc_raw_data["zTotTran"],
            trade_value=tsetmc_raw_data["qTotCap"],
            trade_volume=tsetmc_raw_data["qTotTran5J"],
            last_trade_datetime=last_trade_datetime
        )


@dataclass
class ClientTypeDaily(realtime.ClientType):
    """Holds historical data of an instrument's client type for a single day"""
    record_date: date = None

    def __init__(self, tsetmc_raw_data):
        realtime.ClientType.__init__(
            self=self,
            legal=realtime.ClientTypeTrade(
                buy=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["buy_N_Count"],
                    volume=tsetmc_raw_data["buy_N_Volume"],
                    value=tsetmc_raw_data["buy_N_Value"],
                ),
                sell=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["sell_N_Count"],
                    volume=tsetmc_raw_data["sell_N_Volume"],
                    value=tsetmc_raw_data["sell_N_Value"],
                )
            ),
            natural=realtime.ClientTypeTrade(
                buy=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["buy_I_Count"],
                    volume=tsetmc_raw_data["buy_I_Volume"],
                    value=tsetmc_raw_data["buy_I_Value"],
                ),
                sell=realtime.ClientTypeTradeQuantity(
                    num=tsetmc_raw_data["sell_I_Count"],
                    volume=tsetmc_raw_data["sell_I_Volume"],
                    value=tsetmc_raw_data["sell_I_Value"],
                )
            )
        )
        rd_date = tsetmc_raw_data["recDate"]
        self.record_date = date(
            year=rd_date // 10000,
            month=rd_date // 100 % 100,
            day=rd_date % 100
        )


@dataclass
class TradeIntraday:
    """Intraday micro trade"""
    price: int = None
    volume: int = None
    index: int = None
    time: time = None
    is_canceled: bool = None

    def __init__(self, tsetmc_raw_data):
        self.price = tsetmc_raw_data["pTran"]
        self.volume = tsetmc_raw_data["qTitTran"]
        self.index = tsetmc_raw_data["nTran"]
        h_even = tsetmc_raw_data["hEven"]
        self.time = time(
            hour=h_even // 10000,
            minute=h_even // 100 % 100,
            second=h_even % 100
        )
        self.is_canceled = bool(tsetmc_raw_data["canceled"])


@dataclass
class PriceAdjustment:
    """
    Price adjustment happens when a company distributes dividens \
    or changes the shareholders' equity.
    """
    date: date = None
    price_before: int = None
    price_after: int = None

    def __init__(self, tsetmc_raw_data):
        self.price_before = tsetmc_raw_data["pClosingNotAdjusted"]
        self.price_after = tsetmc_raw_data["pClosing"]
        d_even = tsetmc_raw_data["dEven"]
        self.date = date(
            year=d_even // 10000,
            month=d_even // 100 % 100,
            day=d_even % 100
        )


@dataclass
class InstrumentShareChange:
    """Changes in the total shareholders' equity of a company"""
    record_date: date = None
    total_shares_before: int = None
    total_shares_after: int = None

    def __init__(self, tsetmc_raw_data):
        self.total_shares_before = tsetmc_raw_data["numberOfShareOld"]
        self.total_shares_after = tsetmc_raw_data["numberOfShareNew"]
        d_even = tsetmc_raw_data["dEven"]
        self.record_date = date(
            year=d_even // 10000,
            month=d_even // 100 % 100,
            day=d_even % 100
        )


@dataclass
class BestLimitsHistoryRow(BestLimitsRow):
    """Holds a single row of best limits from an instrument's history"""
    row_number: int = None
    record_time: time = None
    reference_id: int = None

    def __init__(self, tsetmc_raw_data, *args):
        self.row_number = tsetmc_raw_data["number"]
        self.reference_id = tsetmc_raw_data["refID"]
        h_even = tsetmc_raw_data["hEven"]
        self.record_time = time(
            hour=h_even // 10000, minute=h_even // 100 % 100, second=h_even % 100
        )
        BestLimitsRow.__init__(
            self=self,
            tsetmc_raw_data=tsetmc_raw_data,
            *args
        )


@dataclass
class IndexDaily:
    """Daily data for a market index"""
    record_date: date = None
    min_value: float = None
    max_value: float = None
    close_value: float = None

    def __init__(self, tsetmc_raw_data):
        self.min_value = tsetmc_raw_data["xNivInuPbMresIbs"]
        self.max_value = tsetmc_raw_data["xNivInuPhMresIbs"]
        self.close_value = tsetmc_raw_data["xNivInuClMresIbs"]
        d_even = tsetmc_raw_data["dEven"]
        self.record_date = date(
            year=d_even // 10000,
            month=d_even // 100 % 100,
            day=d_even % 100
        )


@dataclass
class InstrumentOptionInfo(
    instrument.ExerciseParams,
    instrument.OptionInstrumentMarginParams
):
    """Technical information of an option contract instrument"""
    tsetmc_code: str = None
    open_positions: int = None
    lot_size: int = None
    begin_date: date = None
    underlying_tsetmc_code: str = None

    def __init__(self, tsetmc_raw_data):
        end_date = tsetmc_raw_data["endDate"]
        exercise_date = date(
            year=end_date // 10000,
            month=end_date // 100 % 100,
            day=end_date % 100
        )
        instrument.ExerciseParams.__init__(
            self=self,
            exercise_price=tsetmc_raw_data["strikePrice"],
            exercise_date=exercise_date
        )
        instrument.OptionInstrumentMarginParams.__init__(
            self=self,
            a_factor=tsetmc_raw_data["aFactor"],
            b_factor=tsetmc_raw_data["bFactor"],
            c_factor=tsetmc_raw_data["cFactor"]
        )
        self.tsetmc_code = tsetmc_raw_data["insCode"]
        self.open_positions = tsetmc_raw_data["buyOP"]
        begin_date = tsetmc_raw_data["beginDate"]
        self.begin_date = date(
            year=begin_date // 10000,
            month=begin_date // 100 % 100,
            day=begin_date % 100
        )
        self.lot_size = tsetmc_raw_data["contractSize"]
        self.underlying_tsetmc_code = tsetmc_raw_data["uaInsCode"]


@dataclass
class IndexValue:
    """Holds the last value of an index with its recent change"""
    close_value: int
    change: int


@dataclass
class PrimaryMarketOverview(realtime.TradeQuantity):
    '''
    Contains data for the main (Bourse) market.
    '''
    overall_index: IndexValue = None
    equal_weighted_index: IndexValue = None
    record_datetime: datetime = None
    market_state: str = None
    market_state_title: str = None
    market_value: int = None

    def __init__(self, tsetmc_raw_data):
        realtime.TradeQuantity.__init__(
            self=self,
            trade_num=tsetmc_raw_data["marketActivityZTotTran"],
            trade_value=tsetmc_raw_data["marketActivityQTotCap"],
            trade_volume=tsetmc_raw_data["marketActivityQTotTran"]
        )
        self.overall_index = IndexValue(
            close_value=tsetmc_raw_data["indexLastValue"],
            change=tsetmc_raw_data["indexChange"]
        )
        self.equal_weighted_index = IndexValue(
            close_value=tsetmc_raw_data["indexEqualWeightedLastValue"],
            change=tsetmc_raw_data["indexEqualWeightedChange"]
        )
        market_activity_deven = tsetmc_raw_data["marketActivityDEven"]
        market_activity_heven = tsetmc_raw_data["marketActivityHEven"]
        self.record_datetime = datetime(
            year=market_activity_deven // 10000, month=market_activity_deven // 100 % 100,
            day=market_activity_deven % 100, hour=market_activity_heven // 10000,
            minute=market_activity_heven // 100 % 100, second=market_activity_heven % 100
        )
        self.market_state = tsetmc_raw_data["marketState"]
        self.market_state_title = tsetmc_raw_data["marketStateTitle"]
        self.market_value = tsetmc_raw_data["marketValue"]


@dataclass
class SecondaryMarketOverview(realtime.TradeQuantity):
    '''
    Contains data for the secondary and tertiary (FarBourse and Paye) markets.
    '''
    secondary_market_index: IndexValue = None
    datetime: datetime = None
    market_state: str = None
    market_state_title: str = None
    market_value: int = None
    tertiary_market_value: int = None

    def __init__(self, tsetmc_raw_data):
        realtime.TradeQuantity.__init__(
            self=self,
            trade_num=tsetmc_raw_data["marketActivityZTotTran"],
            trade_value=tsetmc_raw_data["marketActivityQTotCap"],
            trade_volume=tsetmc_raw_data["marketActivityQTotTran"]
        )
        self.secondary_market_index = IndexValue(
            close_value=tsetmc_raw_data["indexLastValue"],
            change=tsetmc_raw_data["indexChange"]
        )
        market_activity_deven = tsetmc_raw_data["marketActivityDEven"]
        market_activity_heven = tsetmc_raw_data["marketActivityHEven"]
        self.datetime = datetime(
            year=market_activity_deven // 10000, month=market_activity_deven // 100 % 100,
            day=market_activity_deven % 100, hour=market_activity_heven // 10000,
            minute=market_activity_heven // 100 % 100, second=market_activity_heven % 100
        )
        self.market_state = tsetmc_raw_data["marketState"]
        self.market_state_title = tsetmc_raw_data["marketStateTitle"]
        self.market_value = tsetmc_raw_data["marketValue"]
        self.tertiary_market_value = tsetmc_raw_data["marketValueBase"]


@dataclass
class MarketWatchBestLimitsRow(realtime.OrderBookRow):
    """Order book row from market watch"""
    row_id: int = None

    def __init__(self, tsetmc_raw_data):
        realtime.OrderBookRow.__init__(
            self=self,
            demand=realtime.OrderBookRowSide(
                num=int(tsetmc_raw_data["zmd"]),
                volume=int(tsetmc_raw_data["qmd"]),
                price=int(tsetmc_raw_data["pmd"])
            ),
            supply=realtime.OrderBookRowSide(
                num=int(tsetmc_raw_data["zmo"]),
                volume=int(tsetmc_raw_data["qmo"]),
                price=int(tsetmc_raw_data["pmo"])
            )
        )
        self.row_id = tsetmc_raw_data["rid"]


@dataclass
class MarketWatchBestLimits:
    """Instrument order book from market watch"""
    rows: list[MarketWatchBestLimitsRow] = None

    def __init__(self, tsetmc_raw_data):
        self.rows = [MarketWatchBestLimitsRow(x) for x in tsetmc_raw_data]


@dataclass
class MarketWatchTradeData(
    realtime.BigQuantityParams,
):
    """Trade and order book data of an instrument from market watch"""
    price_thresholds: realtime.PriceRange = None
    identification: instrument.InstrumentIdentification = None
    intraday_trade_candle: realtime.TradeCandle = None
    last_trade_time: time = None
    eps: int = None
    orderbook: MarketWatchBestLimits = None

    def __init__(self, tsetmc_raw_data):
        realtime.BigQuantityParams.__init__(
            self=self,
            total_shares=int(tsetmc_raw_data["ztd"]),
            base_volume=int(tsetmc_raw_data["bv"])
        )
        self.price_thresholds = realtime.PriceRange(
            max_price=int(tsetmc_raw_data["pMax"]),
            min_price=int(tsetmc_raw_data["pMin"])
        )
        self.identification = instrument.InstrumentIdentification(
            tsetmc_code=tsetmc_raw_data["insCode"],
            ticker=tsetmc_raw_data["lva"],
            isin=tsetmc_raw_data["insID"],
            name_persian=tsetmc_raw_data["lvc"]
        )
        self.intraday_trade_candle = realtime.TradeCandle(
            previous_price=int(tsetmc_raw_data["py"]),
            last_price=int(tsetmc_raw_data["pdv"]),
            open_price=int(tsetmc_raw_data["pf"]),
            close_price=int(tsetmc_raw_data["pcl"]),
            max_price=int(tsetmc_raw_data["pmx"]),
            min_price=int(tsetmc_raw_data["pmn"]),
            trade_value=int(tsetmc_raw_data["qtc"]),
            trade_volume=int(tsetmc_raw_data["qtj"]),
            trade_num=int(tsetmc_raw_data["ztt"])
        )
        self.eps = tsetmc_raw_data["eps"]
        h_even = tsetmc_raw_data["hEven"]
        self.last_trade_time = time(
            hour=h_even//10000,
            minute=h_even//100 % 100,
            second=h_even % 100
        )
        self.orderbook = MarketWatchBestLimits(
            tsetmc_raw_data=tsetmc_raw_data["blDs"]
        )


@dataclass
class MarketWatchClientTypeData(ClientType):
    """Client type data from market watch"""
    tsetmc_code: str = None

    def __init__(self, tsetmc_raw_data):
        self.tsetmc_code = tsetmc_raw_data["insCode"]
        ClientType.__init__(
            self=self,
            tsetmc_raw_data=tsetmc_raw_data
        )


class TsetmcScrapeException(Exception):
    """Tsetmc bad response status error"""

    def __init__(self, *args, **kwargs):
        self.status_code = kwargs.get('status_code')
        super().__init__(*args)


def ticker_with_tsetmc_homepage_link(
    ticker: str,
    tsetmc_code: str
):
    """
    Returns a HTML element that contains the link to instrument TSETMC homepage
    """
    # pylint: disable = consider-using-f-string
    # Not using f-strings because the VS code autoformatter adds spaces
    return "<a href=\"http://www.tsetmc.com/instInfo/{}\">{}</a>".format(
        tsetmc_code,
        ticker
    )



"""Test tsetmc and tse_client modules in tse_utils library"""
import unittest
from datetime import datetime, date, time
from time import sleep
import httpx
from tse_utils.tsetmc import TsetmcScraper
from tse_utils.tse_client import TseClientScraper
from tse_utils.models import instrument


class TestTSETMC(unittest.IsolatedAsyncioTestCase):
    """Test tsetmc and tse_client modules in tse_utils library"""
    retries_on_timeout = 5

    def __init__(self, *args, **kwargs):
        self.sample_instrument = instrument.Instrument(
            instrument.InstrumentIdentification(
                isin="IRO1FOLD0001", tsetmc_code="46348559193224090",
                ticker="فولاد"
            ))
        self.sample_date = date(year=2023, month=4, day=30)
        self.sample_index_identification = instrument.IndexIdentification(
            persian_name="شاخص کل", tsetmc_code="32097828799138957")
        self.sample_option = instrument.OptionInstrument(
            exercise_date=date(year=2023, month=10, day=18),
            exercise_price=1653,
            underlying=self.sample_instrument,
            identification=instrument.InstrumentIdentification(
                isin="IRO9FOLD6821",
                ticker="ضفلا7020",
                tsetmc_code="37762443198265540"
            )
        )
        super().__init__(*args, **kwargs)

    async def test_get_instrument_identity(self):
        """Test instrument identity fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    identity = await tsetmc.get_instrument_identity(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertEqual(
                        identity.isin, self.sample_instrument.identification.isin)
            except httpx.ConnectTimeout as exc:
                if trial_ind == self.retries_on_timeout - 1:
                    raise exc
                sleep(1)

    async def test_get_instrument_search(self):
        """Test instrument search on TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    search_result = await tsetmc.get_instrument_search(
                        self.sample_instrument.identification.ticker
                    )
                    self.assertTrue(any(
                        x.tsetmc_code == self.sample_instrument.identification.tsetmc_code
                        for x in search_result
                    ))
            except httpx.ConnectTimeout as exc:
                if trial_ind == self.retries_on_timeout - 1:
                    raise exc
                sleep(1)

    async def test_get_closing_price_info(self):
        """Test instrument current trade data fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_closing_price_info(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(not data.last_trade_datetime is None)
                    self.assertTrue(data.min_price *
                                    data.trade_volume <= data.trade_value)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_instrument_info(self):
        """Test instrument home page data fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_instrument_info(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(
                        data.isin == self.sample_instrument.identification.isin)
                    self.assertTrue(data.total_shares == 800000000000)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_client_type(self):
        """Test instrument current client type fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_client_type(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(
                        data.legal.sell.volume + data.natural.sell.volume ==
                        data.legal.buy.volume + data.natural.buy.volume
                    )
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_best_limits(self):
        """Test instrument current order book fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_best_limits(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(len(data.rows) == 5)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_closing_price_daily_list(self):
        """Test instrument historical daily prices fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_closing_price_daily_list(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(len(data) > 1000)
                    chosen_date = datetime(year=2023, month=9, day=11)
                    date_data = next(
                        x for x in data if x.last_trade_datetime.date() == chosen_date.date())
                    self.assertTrue(date_data.trade_volume == 74309985)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_client_type_history(self):
        """Test instrument historical daily client type fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_client_type_daily_list(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    chosen_date = date(year=2023, month=9, day=11)
                    date_data = next(
                        x for x in data if x.record_date == chosen_date)
                    self.assertTrue(date_data.legal.buy.num == 13)
                    self.assertTrue(date_data.legal.buy.value == 259393764720)
                    self.assertTrue(date_data.legal.buy.volume == 46754064)
                    self.assertTrue(date_data.legal.sell.num == 15)
                    self.assertTrue(date_data.legal.sell.value == 192455986390)
                    self.assertTrue(date_data.legal.sell.volume == 34727215)
                    self.assertTrue(date_data.natural.buy.num == 1277)
                    self.assertTrue(
                        date_data.natural.buy.value == 152695292260)
                    self.assertTrue(date_data.natural.buy.volume == 27555921)
                    self.assertTrue(date_data.natural.sell.num == 1242)
                    self.assertTrue(
                        date_data.natural.sell.value == 219633070590)
                    self.assertTrue(date_data.natural.sell.volume == 39582770)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_trade_intraday_list(self):
        """Test instrument intradary microtrades fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_trade_intraday_list(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(len(data) != 0)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_price_adjustment_list(self):
        """Test instrument historical price adjustments fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_price_adjustment_list(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(len(data) > 10)
                    self.assertTrue(any(x.date == date(year=2023, month=7, day=22)
                                        and x.price_before == 5460
                                        and x.price_after == 4960 for x in data))
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_instrument_share_change(self):
        """Test instrument historical equity change fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_instrument_share_change(
                        self.sample_instrument.identification.tsetmc_code
                    )
                    self.assertTrue(len(data) > 5)
                    self.assertTrue(any(
                        x.record_date == date(year=2022, month=8, day=9)
                        and x.total_shares_before == 293000000000
                        and x.total_shares_after == 530000000000
                        for x in data
                    ))
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_trade_intraday_hisory_list(self):
        """Test instrument historical intraday microtrades fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_trade_intraday_hisory_list(
                        self.sample_instrument.identification.tsetmc_code, self.sample_date)
                    chosen_data = next(x for x in data if x.index == 15552)
                    self.assertTrue(chosen_data.volume == 100790
                                    and chosen_data.price == 7250
                                    and chosen_data.time == time(hour=12, minute=26, second=6)
                                    and chosen_data.is_canceled)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_best_limits_intraday_history_list(self):
        """Test instrument historical orderbook fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_best_limits_intraday_history_list(
                        self.sample_instrument.identification.tsetmc_code, self.sample_date
                    )
                    self.assertTrue(
                        any(x.record_time == time(hour=8, minute=45, second=36) and
                            x.row_number == 5 and x.reference_id == 11679170214 and
                            x.demand.volume == 163213 and x.demand.price == 7000
                            for x in data
                            )
                    )
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_index_history(self):
        """Test historical index data fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_index_history(
                        self.sample_index_identification.tsetmc_code
                    )
                    self.assertTrue(
                        any(
                            x.record_date == date(year=2023, month=9, day=13) and
                            x.close_value == 2126741.7 and x.min_value == 2126690 and
                            x.max_value == 2130510
                            for x in data
                        )
                    )
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_instrument_option_info(self):
        """Test instrument option info fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_instrument_option_info(
                        self.sample_option.identification.isin
                    )
                    self.assertTrue(
                        data.exercise_date ==
                        self.sample_option.exercise_date and
                        data.exercise_price ==
                        self.sample_option.exercise_price and
                        data.underlying_tsetmc_code ==
                        self.sample_instrument.identification.tsetmc_code
                    )
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_primary_market_overview(self):
        """Test primary market overview fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_primary_market_overview()
                    self.assertTrue(data.market_value > 4e16)
                    self.assertTrue(data.overall_index.close_value > 1e6)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_secondary_market_overview(self):
        """Test secondary market overview fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_secondary_market_overview()
                    self.assertTrue(data.market_value > 1e16)
                    self.assertTrue(
                        data.secondary_market_index.close_value > 1e4)
                    self.assertTrue(data.tertiary_market_value > 1e15)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_market_watch(self):
        """Test market watch fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_market_watch()
                    self.assertTrue(len(data) > 100)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_get_client_type_all(self):
        """Test market client type fetch from TSETMC"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TsetmcScraper() as tsetmc:
                    data = await tsetmc.get_client_type_all()
                    self.assertTrue(len(data) > 100)
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)

    async def test_tse_client_get_instruments_list(self):
        """Test fetch instruments list from TseClient"""
        for trial_ind in range(self.retries_on_timeout):
            try:
                async with TseClientScraper() as tse_client:
                    instruments, _ = await tse_client.get_instruments_list()
                    self.assertTrue(len(instruments) > 0)
                    self.assertTrue(
                        any(
                            x.tsetmc_code == self.sample_instrument.identification.tsetmc_code and
                            x.isin == self.sample_instrument.identification.isin
                            for x in instruments
                        )
                    )
            except httpx.ConnectTimeout:
                if trial_ind == self.retries_on_timeout - 1:
                    raise
                sleep(1)


if __name__ == '__main__':
    unittest.main()



"""
This module holds tse client modules.
"""
from dataclasses import dataclass
from datetime import date
from tse_utils.models import instrument


@dataclass
class TseClientInstrumentIdentitification(instrument.InstrumentIdentification):
    """Instrument Identification from TseClient"""
    last_change_date: date = None
    is_index: bool = None
    market_code: int = None
    sector_code: int = None
    sub_sector_code: int = None
    type_id: int = None

    def __init__(self, tseclient_raw_data):
        instrument.InstrumentIdentification.__init__(
            self=self,
            tsetmc_code=tseclient_raw_data[0],
            isin=tseclient_raw_data[1],
            ticker=tseclient_raw_data[5],
            name_persian=tseclient_raw_data[6],
            name_english=tseclient_raw_data[3]
        )
        self.type_id = int(tseclient_raw_data[17])
        sector_code = str(tseclient_raw_data[15]).replace(" ", "")
        if sector_code.isdigit():
            self.sector_code = int(sector_code)
            self.sub_sector_code = int(tseclient_raw_data[16])
            self.is_index = False
        else:
            self.is_index = True
        self.market_code = int(tseclient_raw_data[9])
        last_change_date_raw = int(tseclient_raw_data[8])
        self.last_change_date = date(
            year=last_change_date_raw//10000,
            month=last_change_date_raw//100 % 100,
            day=last_change_date_raw % 100
        )


class TseClientScrapeException(Exception):
    """Tse client bad response status error"""

    def __init__(self, *args, **kwargs):
        self.status_code = kwargs.get('status_code')
        super().__init__(*args)



"""
Realtime data for instruments are of different types
Classes in the realtime module holds such data
"""
from dataclasses import dataclass
from datetime import datetime
import threading
from tse_utils.models.enums import Nsc


@dataclass
class OrderBookRowSide:
    """A single side on a row of an instrument's order book"""
    num: int = 0
    volume: int = 0
    price: int = 0

    def __str__(self) -> str:
        return f"{self.volume} @ {self.price}"

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, OrderBookRowSide) and
            __value.num == self.num and
            __value.volume == self.volume and
            __value.price == self.price
        )


@dataclass
class OrderBookRow():
    """
    Contains a single row from an instrument's order book
    """
    demand: OrderBookRowSide
    supply: OrderBookRowSide

    def __init__(
            self,
            demand: OrderBookRowSide = None,
            supply: OrderBookRowSide = None
    ):
        self.demand = demand if demand else OrderBookRowSide()
        self.supply = supply if supply else OrderBookRowSide()

    def __str__(self):
        return f"{self.supply.num} {self.supply.volume} {self.supply.price} | \
            {self.demand.price} {self.demand.volume} {self.demand.num}"

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, OrderBookRow) and
            __value.demand == self.demand and
            __value.supply == self.supply
        )


@dataclass
class OrderBook():
    """
    Orderbook contains the top n rows of an instrument's orders on both sides.
    """

    def __init__(self, row_count: int = 5):
        self.rows: list[OrderBookRow] = [OrderBookRow() for i in range(row_count)]

    # def get_diff(self): TODO: get difference with order


@dataclass
class ClientTypeTradeQuantity:
    """Holds trade quantity for a single side of a single type of client"""
    num: int = None
    volume: int = None
    # In many cases the value parameter is not valued
    value: int = None

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, ClientTypeTradeQuantity) and
            __value.num == self.num and
            __value.volume == self.volume
        )


@dataclass
class ClientTypeTrade:
    """Holds trades for a single type of client"""
    buy: ClientTypeTradeQuantity
    sell: ClientTypeTradeQuantity

    def __init__(
            self,
            buy: ClientTypeTradeQuantity = None,
            sell: ClientTypeTradeQuantity = None
    ):
        self.buy = buy if buy else ClientTypeTradeQuantity()
        self.sell = sell if sell else ClientTypeTradeQuantity()

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, ClientTypeTrade) and
            __value.buy == self.buy and
            __value.sell == self.sell
        )


@dataclass
class ClientType:
    """
    ClientType holds data used to separate the share of the natural investors \
    from the legal ones on trades
    """
    legal: ClientTypeTrade
    natural: ClientTypeTrade

    def __init__(
            self,
            legal: ClientTypeTrade = None,
            natural: ClientTypeTrade = None
    ):
        self.legal = legal if legal else ClientTypeTrade()
        self.natural = natural if natural else ClientTypeTrade()

    def trade_volume(self) -> int:
        """returns total trade volume"""
        return self.legal.buy.volume + self.natural.buy.volume

    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, ClientType) and
            __value.legal == self.legal and
            __value.natural == self.natural
        )


@dataclass
class PriceRange:
    """Holds price thresholds"""
    max_price: int = None
    min_price: int = None


@dataclass
class TradeQuantity:
    """Holds trade quantity identifiers"""
    trade_num: int = None
    trade_value: int = None
    trade_volume: int = None


@dataclass
class TradeCandle(PriceRange, TradeQuantity):
    """Holds current or point in time trade data for an instrument"""
    previous_price: int = None
    open_price: int = None
    close_price: int = None
    last_price: int = None
    open_trade_datetime: datetime = None
    last_trade_datetime: datetime = None


class DeepOrderBook:
    """
    DeepOrderbook contains all rows of an instrument's orders on both sides.
    """

    def __init__(self):
        self._buy_rows: list[OrderBookRowSide] = []
        self._buy_rows_lock: threading.Lock = threading.Lock()
        self._sell_rows: list[OrderBookRowSide] = []
        self._sell_rows_lock: threading.Lock = threading.Lock()

    def update_buy_row(self, num: int, volume: int, price: int) -> None:
        """Updates a single buy row if exists and adds it if not"""
        with self._buy_rows_lock:
            row = next((x for x in self._buy_rows if x.price == price), None)
            if row:
                row.volume = volume
                row.num = num
            else:
                self._buy_rows.append(OrderBookRowSide(
                    num=num, volume=volume, price=price))

    def update_sell_row(self, num: int, volume: int, price: int) -> None:
        """Updates a single sell row if exists and adds it if not"""
        with self._sell_rows_lock:
            row = next((x for x in self._sell_rows if x.price == price), None)
            if row:
                row.volume = volume
                row.num = num
            else:
                self._sell_rows.append(OrderBookRowSide(
                    num=num, volume=volume, price=price))

    def remove_buy_row(self, price: int) -> None:
        """Removes a single buy row"""
        with self._buy_rows_lock:
            row = next((x for x in self._buy_rows if x.price == price), None)
            if row:
                self._buy_rows.remove(row)

    def remove_sell_row(self, price: int) -> None:
        """Removes a single sell row"""
        with self._sell_rows_lock:
            row = next((x for x in self._sell_rows if x.price == price), None)
            if row:
                self._sell_rows.remove(row)

    def empty_buy_rows(self) -> None:
        """Removes all buy rows"""
        with self._buy_rows_lock:
            self._buy_rows.clear()

    def empty_sell_rows(self) -> None:
        """Removes all sell rows"""
        with self._sell_rows_lock:
            self._sell_rows.clear()

    def get_buy_rows(self) -> list[OrderBookRowSide]:
        """Returns a copy of all buy rows"""
        with self._buy_rows_lock:
            return sorted(self._buy_rows.copy(), key=lambda x: x.price, reverse=True)

    def get_sell_rows(self) -> list[OrderBookRowSide]:
        """Returns a copy of all sell rows"""
        with self._sell_rows_lock:
            return sorted(self._sell_rows.copy(), key=lambda x: x.price, reverse=False)


@dataclass
class OrderPriceLimitations(PriceRange):
    """Holds limitations on prices that mostly change daily"""
    price_tick: int = 1


@dataclass
class OrderQuantityLimitations:
    """Holds limitations on quantity of orders"""
    max_buy_order_quantity: int = None
    max_sell_order_quantity: int = None
    lot_size: int = 1


@dataclass
class OrderLimitations(
    OrderPriceLimitations,
    OrderQuantityLimitations
):
    """Holds the daily limitations for trading on instrument"""
    nsc: Nsc = None


@dataclass
class BigQuantityParams:
    """Holds big quantity parameters for instrument"""
    base_volume: int = 1
    total_shares: int = None



"""
This model contains classes that can be used as interfaces for \
implementing the trader classes in the future. Each trader instance \
is responsible for a single account in a specific broker and OMS.
"""
from dataclasses import dataclass
from datetime import date, datetime
import threading
import logging
from abc import ABC, abstractmethod
from typing import Callable
from tse_utils.models.enums import (
    TradeSide,
    TraderConnectionState,
    OrderLock,
    OrderState,
    OrderValidityType
)
from tse_utils.models import instrument


@dataclass
class MicroTrade:
    """
    Holds data for a single micro trade.
    """
    id = None
    isin: str = None
    side: TradeSide = None
    quantity: int = None
    price: int = None
    datetime: datetime = None
    htn: str = None


@dataclass
class OrderIdentifier:
    """Holds the identifiers of an Order"""
    oms_id: str | int
    isin: str
    side: TradeSide
    client_id: str = None
    """
    client_id is used for mapping the order on the OMS side with the one sent by a client
    """
    hon: str = None
    """
    hon identifies the position of order in the market kernel
    """


@dataclass
class OrderStatus:
    """Contains both OrderState and OrderLock"""
    state: OrderState = None
    lock: OrderLock = None


@dataclass
class OrderQuantity:
    """Holds all quantity information for an order"""
    quantity: int
    remaining_quantity: int = None
    executed_quantity: int = None


@dataclass
class OrderValidity:
    """Contains validity type of order and expiration date, if appropriate"""
    validity_type: OrderValidityType = None
    creation_datetime: datetime = None
    expiration_date: date = None
    """
    expiration_date is used for orders validity of which are GOOD_TILL_DATE
    """


class Order(OrderIdentifier, OrderStatus, OrderQuantity, OrderValidity):
    """
    Holds data for a single order from a trader.
    """

    # pylint: disable=too-many-arguments
    # The following 6 parameters are the bare minimums \
    # with which an order is identified
    def __init__(
        self,
        oms_id,
        isin: str,
        side: TradeSide,
        quantity: int,
        price: int
    ):
        OrderIdentifier.__init__(
            self=self,
            oms_id=oms_id,
            side=side,
            isin=isin
        )
        OrderStatus.__init__(self=self)
        OrderQuantity.__init__(
            self=self,
            quantity=quantity
        )
        OrderValidity.__init__(self=self)
        self.price: int = price
        self.blocked_credit: int = None
        self._trades: list[MicroTrade] = []
        self._trades_lock: threading.Lock = threading.Lock()

    def add_trade(self, trade: MicroTrade) -> None:
        """Add new trade to the list of order trades"""
        with self._trades_lock:
            self._trades.append(trade)

    def get_trades(self) -> list[MicroTrade]:
        """Get a copy of the order's trades list"""
        with self._trades_lock:
            return self._trades.copy()

    def __str__(self) -> str:
        return f"{self.side}|{self.state}|{self.oms_id}|\
            {self.isin}|P:{self.price},Q:{self.quantity}"


@dataclass
class PortfolioCash:
    """Holds the cash and credit in an account's portfolio"""
    free_balance: int = None
    blocked_balance: int = None
    accounting_remain: int = None
    credit_limit: int = None

    def get_all_balance(self) -> int:
        """Get net balance of the account (without credit)"""
        return self.free_balance + self.blocked_balance


@dataclass
class PortfolioSecurity:
    """Holds data for a single security asset in the porfolio"""
    isin: str
    """
    positive quantity means a long position and a negative ones means a short position.
    """
    quantity: int = None
    position_open_price: int = None
    instrument_last_price: int = None
    instrument_close_price: int = None

    def __str__(self) -> str:
        return f"{self.quantity} of {self.isin}"


class Portfolio:
    """A trader account's portfolio consisting of cash, securities, and positions"""

    def __init__(self):
        self.cash: PortfolioCash = PortfolioCash()
        """
        The assets list includes instruments in which having a net short position is unallowed.
        """
        self._assets: list[PortfolioSecurity] = []
        self._assets_lock = threading.Lock()
        """
        The positions list includes instruments in which \
        having a net short position is possible, such as options and futures.
        """
        self._positions: list[PortfolioSecurity] = []
        self._positions_lock = threading.Lock()

    def has_asset(self, isin: str) -> bool:
        """Checks if portfolio has any asset of a specific security"""
        with self._assets_lock:
            return any(x for x in self._assets if x.isin == isin)

    def get_asset(self, isin: str) -> PortfolioSecurity:
        """Get asset in the portfolio from a specific security"""
        with self._assets_lock:
            return next((x for x in self._assets if x.isin == isin), None)

    def get_asset_quantity(self, isin: str) -> int:
        """Get asset quantity in the portfolio from a specific security"""
        with self._assets_lock:
            return next((x.quantity for x in self._assets if x.isin == isin), 0)

    def get_all_assets(self) -> list[PortfolioSecurity]:
        """Get a copy of the assets list in the portfolio"""
        with self._assets_lock:
            return self._assets.copy()

    def remove_asset(self, isin: str) -> None:
        """Remove an asset from the portfolio"""
        with self._assets_lock:
            asset = next((x for x in self._assets if x.isin == isin), None)
            if asset:
                self._assets.remove(asset)

    def empty_asset(self) -> None:
        """Remove all assets from the portfolio"""
        with self._assets_lock:
            self._assets.clear()

    def update_asset(
            self,
            security: PortfolioSecurity
    ) -> None:
        """Updates a specific asset in the portfolio"""
        with self._assets_lock:
            asset = next(
                (x for x in self._assets if x.isin == security.isin), None)
            if asset:
                asset.quantity = security.quantity
                asset.position_open_price = security.position_open_price
                asset.instrument_close_price = security.instrument_close_price
                asset.instrument_last_price = security.instrument_last_price
            else:
                self._assets.append(
                    PortfolioSecurity(
                        isin=security.isin,
                        quantity=security.quantity,
                        position_open_price=security.position_open_price,
                        instrument_close_price=security.instrument_close_price,
                        instrument_last_price=security.instrument_last_price
                    ))

    def has_position(self, isin: str) -> bool:
        """Checks if portfolio has any position of a specific security"""
        with self._positions_lock:
            return any(
                x
                for x in self._positions
                if x.isin == isin
            )

    def get_position(self, isin: str) -> PortfolioSecurity:
        """Get position in the portfolio from a specific security"""
        with self._positions_lock:
            return next((x for x in self._positions if x.isin == isin), None)

    def get_position_quantity(self, isin: str) -> int:
        """Get position quantity in the portfolio from a specific security"""
        with self._positions_lock:
            return next((x.quantity for x in self._positions if x.isin == isin), 0)

    def get_all_positions(self) -> list[PortfolioSecurity]:
        """Get a copy of the positions list in the portfolio"""
        with self._positions_lock:
            return self._positions.copy()

    def remove_position(self, isin: str) -> None:
        """Remove a position from the portfolio"""
        with self._positions_lock:
            position = next(
                (x for x in self._positions if x.isin == isin), None)
            if position:
                self._positions.remove(position)

    def empty_position(self) -> None:
        """Remove all positions from the portfolio"""
        with self._positions_lock:
            self._positions.clear()

    def update_position(
            self,
            security: PortfolioSecurity
    ) -> None:
        """Updates a specific position in the portfolio"""
        with self._positions_lock:
            position = next(
                (x for x in self._positions if x.isin == security.isin), None)
            if position:
                position.quantity = security.quantity
                position.position_open_price = security.position_open_price
                position.instrument_close_price = security.instrument_close_price
                position.instrument_last_price = security.instrument_last_price
            else:
                self._positions.append(
                    PortfolioSecurity(
                        isin=security.isin,
                        quantity=security.quantity,
                        position_open_price=security.position_open_price,
                        instrument_close_price=security.instrument_close_price,
                        instrument_last_price=security.instrument_last_price
                    ))


@dataclass
class TradingAPI:
    """Each trading API consists of a single OMS and a single broker"""
    broker_title: str = None
    oms_title: str = None
    oms_domain: str = None

    def __str__(self) -> str:
        return f"{self.oms_title} - {self.broker_title}"


@dataclass
class TraderCredentials:
    """Holds the credentials of a single trader account"""
    api: TradingAPI
    username: str
    # password is optional since some APIs work with long-lived tokens
    password: str = None


@dataclass
class TraderIdentification:
    """Holds the identification of a single trader account"""
    dbid: int = None
    first_name: str = None
    last_name: str = None
    """
    display_name is used for logging purposes. If it is not set, username will be used.
    """
    display_name: str = None
    bourse_code: str = None
    oms_code: str = None

    def __str__(self) -> str:
        return self.display_name


@dataclass
class TraderRealtimeData:
    """
    Contains the realtime data for a single trader account.
    This data is mostly pushed through subscriptions to websockets.
    """
    portfolio: Portfolio
    _orders: list[Order]
    _orders_lock: threading.Lock
    _subscribed_instruments: list[instrument.Instrument]

    def __init__(self):
        self.portfolio: Portfolio = Portfolio()
        self._orders: list[Order] = []
        self._orders_lock: threading.Lock = threading.Lock()
        self._subscribed_instruments: list[instrument.Instrument] = []


class Trader(ABC, TraderRealtimeData):
    """
    Trader class holds the data for a single trader account \
    and can be inheridated by classes specialized in training \
    using a specific OMS API.
    """

    def __init__(
            self,
            credentials: TraderCredentials,
            logger_name: str = None
    ):
        self.credentials: TraderCredentials = credentials
        self.identification: TraderIdentification = TraderIdentification()
        self.logger: logging.Logger = logging.getLogger(logger_name)
        self.connection_state: TraderConnectionState = TraderConnectionState.NO_LOGIN
        TraderRealtimeData.__init__(self=self)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        try:
            await self.disconnect()
        except Exception as ex:
            self.logger.error(
                "Disconnect failed on disposing object."
            )
            self.logger.debug(
                "Encountered %s when trying to disconnect.",
                ex
            )
            raise ex

    def __str__(self) -> str:
        name = self.identification.display_name \
            if self.identification.display_name \
            else self.credentials.username
        return f"{name}@{self.credentials.api.broker_title}"

    def get_order(self, oms_id) -> Order:
        """Searchs for an order in the orders list using its oms_id"""
        with self._orders_lock:
            return next((x for x in self._orders if x.oms_id == oms_id), None)

    def get_order_custom(
            self,
            filter_func: Callable[[Order], bool]
    ) -> Order:
        """Searchs for an order in the orders list using a custom filter"""
        with self._orders_lock:
            return next((x for x in self._orders if filter_func(x)), None)

    def get_orders(
            self,
            filter_func: Callable[[Order], bool] = lambda x: True
    ) -> list[Order]:
        """Searchs for all orders complying with a filter in the orders list"""
        with self._orders_lock:
            return [x for x in self._orders if filter_func(x)]

    def remove_order(self, oms_id) -> None:
        """Removes an order from the orders list using its oms_id"""
        with self._orders_lock:
            order = next((x for x in self._orders if x.oms_id == oms_id), None)
            if order:
                self._orders.remove(order)

    def empty_orders(self) -> None:
        """Removes all orders from the orders list"""
        with self._orders_lock:
            self._orders.clear()

    def add_order(self, order: Order) -> None:
        """
        Adds a new order to the orders list.
        This method should be used by data pushers and not by the client.
        Client can use order_send to send a new order.
        """
        with self._orders_lock:
            self._orders.append(order)

    def get_subscribed_instrument(self, isin: str = None):
        """Gets a subscribed instrument from the subscribed instruments list"""
        return next((
            x
            for x in self._subscribed_instruments
            if x.identification.isin == isin
        ), None)

    @abstractmethod
    async def connect(self) -> None:
        """
        Does a single attempt on logging in to the trader account
        """

    @abstractmethod
    async def connect_looper(self, interval: int = 3, max_trial=10) -> None:
        """
        Recursively trying to login to the trader account
        """

    @abstractmethod
    async def disconnect(self) -> None:
        """
        Logs of from the trader account
        """

    @abstractmethod
    async def get_server_datetime(self) -> datetime:
        """
        Gets server datetime
        """

    @abstractmethod
    async def pull_trader_data(self):
        """
        Fetches client data from OMS using a pull request
        """

    @abstractmethod
    async def subscribe_instruments_list(
        self,
        instruments: list[instrument.Instrument]
    ):
        """
        Subscribes the instruments on the OMS pusher to get their realtime data
        """

    @abstractmethod
    async def order_send(
        self,
        order: Order
    ):
        """
        Used for sending a new order to the OMS. 
        The input order object should be used as a data transfer object \
        and is not added to the orders list, since new orders to be added \
        there should come from the OMS pushers.
        """

    @abstractmethod
    async def order_cancel(self, order: Order):
        """
        Used for canceling an active order
        """

    @abstractmethod
    async def order_edit(self, order: Order, quantity: int, price: int):
        """
        Used for editing an active order
        """



"""Enums used in tse-utils are in the enums module"""
from enum import Enum


class Nsc(Enum):
    """Nsc state of instruments"""
    A = "مجاز"
    AG = "مجاز مسدود"
    AR = "مجاز محفوظ"
    AS = "مجاز متوقف"
    I = "ممنوع"
    IG = "ممنوع مسدود"
    IR = "ممنوع محفوظ"
    IS = "ممنوع متوقف"


class TradeSide(Enum):
    """Trade/Order side"""
    BUY = "خرید"
    SELL = "فروش"


class OrderState(Enum):
    """Trader orders' state"""
    SENT_TO_CORE = "ارسال به هسته"
    ACTIVE = "فعال"
    EXECUTED = "اجرا شده"
    CANCELED = "حذف شده"
    ERROR = "خطا"

    def is_active(self) -> bool:
        """Checks if order is active"""
        return self == OrderState.ACTIVE


class OrderValidityType(Enum):
    """Validity type for trader orders"""
    DAY = "روز"
    GOOD_TILL_CANCEL = "معتبر تا لغو"
    GOOD_TILL_DATE = "معتبر تا تاریخ"
    FILL_OR_KILL = "اجرا و حذف"


class OrderLock(Enum):
    """Trader orders' lock state"""
    UNLOCK = "آزاد"
    LOCK_FOR_CREATION = "قفل برای ایجاد"
    LOCK_FOR_EDITION = "قفل برای ویرایش"
    LOCK_FOR_CANCELATION = "قفل برای حذف"


class TraderConnectionState(Enum):
    """Connection state of trader account"""
    NO_LOGIN = "وارد نشده"
    CONNECTING = "در حال اتصال"
    CONNECTED = "متصل"
    RECONNECTING = "در حال اتصال مجدد"
    CONNECTION_BROKEN = "قطع اتصال"
    LOGGED_OUT = "خارج شده"

    def is_stable(self) -> bool:
        """Checks if connection is in a stable status"""
        return self in (
            TraderConnectionState.NO_LOGIN,
            TraderConnectionState.CONNECTED,
            TraderConnectionState.LOGGED_OUT
        )

    def can_request_connect(self) -> bool:
        """Checks if connection is down and stable"""
        return self in (
            TraderConnectionState.NO_LOGIN,
            TraderConnectionState.LOGGED_OUT
        )



"""
This module uses httpx to fetch data asynchronously \
from the TSETMC website. 
"""
from datetime import date
import json
import httpx
from tse_utils.tsetmc.models import (
    TsetmcScrapeException,
    InstrumentIdentification,
    InstrumentSearchItem,
    ClosingPriceInfo,
    InstrumentInfo,
    ClientType,
    BestLimits,
    ClosingPriceDaily,
    ClientTypeDaily,
    TradeIntraday,
    PriceAdjustment,
    InstrumentShareChange,
    BestLimitsHistoryRow,
    IndexDaily,
    PrimaryMarketOverview,
    InstrumentOptionInfo,
    SecondaryMarketOverview,
    MarketWatchTradeData,
    MarketWatchClientTypeData
)


class TsetmcScraper():
    """
    This class fetches data from tsetmc.com, the official website 
    for Tehran Stock Exchange market data.
    """

    def __init__(self, tsetmc_domain: str = "cdn.tsetmc.com"):
        self.tsetmc_domain = tsetmc_domain
        self.__client = httpx.AsyncClient(headers={
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) \
                AppleWebKit/537.36 (KHTML, like Gecko) \
                Chrome/89.0.4389.114 Safari/537.36",
            "accept": "application/json, text/plain, */*"
        }, base_url=f"https://{tsetmc_domain}/")

    async def __aenter__(self):
        return self

    async def __aexit__(
            self,
            exc_type,
            exc_value,
            traceback
    ):
        await self.__client.aclose()

    async def __get_instrument_identity_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument identity card"""
        req = await self.__client.get(
            f"api/Instrument/GetInstrumentIdentity/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_instrument_identity(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> InstrumentIdentification:
        """Get processed instrument identity card"""
        raw = await self.__get_instrument_identity_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return InstrumentIdentification(
            tsetmc_code=tsetmc_code,
            tsetmc_raw_data=raw["instrumentIdentity"]
        )

    async def __get_instrument_search_raw(
            self,
            search_value: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument search results"""
        req = await self.__client.get(
            f"api/Instrument/GetInstrumentSearch/{search_value}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_instrument_search(
            self,
            search_value: str,
            timeout: int = 3
    ) -> list[InstrumentSearchItem]:
        """Get and process instrument search results"""
        raw = await self.__get_instrument_search_raw(
            search_value=search_value,
            timeout=timeout
        )
        return [InstrumentSearchItem(x) for x in raw["instrumentSearch"]]

    async def __get_closing_price_info_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument current trade data"""
        req = await self.__client.get(
            f"api/ClosingPrice/GetClosingPriceInfo/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_closing_price_info(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> ClosingPriceInfo:
        """Get and process instrument current trade data"""
        raw = await self.__get_closing_price_info_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return ClosingPriceInfo(tsetmc_raw_data=raw["closingPriceInfo"])

    async def __get_instrument_info_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument home page data"""
        req = await self.__client.get(
            f"api/Instrument/GetInstrumentInfo/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_instrument_info(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> InstrumentInfo:
        """Get and process instrument home page data"""
        raw = await self.__get_instrument_info_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return InstrumentInfo(
            tsetmc_code=tsetmc_code,
            tsetmc_raw_data=raw["instrumentInfo"]
        )

    async def __get_client_type_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument current client type data"""
        req = await self.__client.get(
            f"api/ClientType/GetClientType/{tsetmc_code}/1/0",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_client_type(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> ClientType:
        """Get and process instrument current client type data"""
        raw = await self.__get_client_type_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return ClientType(tsetmc_raw_data=raw["clientType"])

    async def __get_best_limits_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument order book data"""
        req = await self.__client.get(
            f"api/BestLimits/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_best_limits(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> BestLimits:
        """Get and process instrument order book data"""
        raw = await self.__get_best_limits_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return BestLimits(tsetmc_raw_data=raw["bestLimits"])

    async def __get_closing_price_daily_list_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument daily historical trade data"""
        req = await self.__client.get(
            f"api/ClosingPrice/GetClosingPriceDailyList/{tsetmc_code}/0",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_closing_price_daily_list(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[ClosingPriceDaily]:
        """Get and process instrument daily historical trade data"""
        raw = await self.__get_closing_price_daily_list_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [
            ClosingPriceDaily(tsetmc_raw_data=x)
            for x in raw["closingPriceDaily"]
        ][::-1]

    async def __get_client_type_daily_list_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument daily historical client type data"""
        req = await self.__client.get(
            f"api/ClientType/GetClientTypeHistory/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_client_type_daily_list(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[ClientTypeDaily]:
        """Get and process instrument daily historical client type data"""
        raw = await self.__get_client_type_daily_list_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [
            ClientTypeDaily(tsetmc_raw_data=x)
            for x in raw["clientType"]
        ][::-1]

    async def __get_trade_intraday_list_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument intraday microtrades data"""
        req = await self.__client.get(
            f"api/Trade/GetTrade/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_trade_intraday_list(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[TradeIntraday]:
        """Get and process instrument intraday microtrades data"""
        raw = await self.__get_trade_intraday_list_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [TradeIntraday(tsetmc_raw_data=x) for x in raw["trade"]]

    async def __get_price_adjustment_list_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument historical price adjustments"""
        req = await self.__client.get(
            f"api/ClosingPrice/GetPriceAdjustList/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_price_adjustment_list(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[PriceAdjustment]:
        """Get and process instrument historical price adjustments"""
        raw = await self.__get_price_adjustment_list_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [PriceAdjustment(tsetmc_raw_data=x) for x in raw["priceAdjust"]]

    async def __get_instrument_share_change_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument historical share changes"""
        req = await self.__client.get(
            f"api/Instrument/GetInstrumentShareChange/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_instrument_share_change(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[InstrumentShareChange]:
        """Get and process instrument historical share changes"""
        raw = await self.__get_instrument_share_change_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [
            InstrumentShareChange(tsetmc_raw_data=x)
            for x in raw["instrumentShareChange"]
        ]

    async def __get_trade_intraday_hisory_list_raw(
            self,
            tsetmc_code: str,
            query_date: date,
            detailed: bool = True,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument historical intraday microtrades"""
        req = await self.__client.get(
            f"api/Trade/GetTradeHistory/{tsetmc_code}/\
                {query_date.year}{query_date.month:02}{query_date.day:02}/\
                    {not detailed}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_trade_intraday_hisory_list(
            self,
            tsetmc_code: str,
            query_date: date,
            detailed: bool = True,
            timeout: int = 3
    ) -> list[TradeIntraday]:
        """Get and process instrument historical intraday microtrades"""
        raw = await self.__get_trade_intraday_hisory_list_raw(
            tsetmc_code=tsetmc_code,
            query_date=query_date,
            detailed=detailed,
            timeout=timeout
        )
        proc = [
            TradeIntraday(tsetmc_raw_data=x)
            for x in raw["tradeHistory"]
        ]
        proc.sort(key=lambda x: x.index)
        return proc

    async def __get_best_limits_intraday_history_list_raw(
            self,
            tsetmc_code: str,
            query_date: date,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument historical intraday order book"""
        req = await self.__client.get(
            f"api/BestLimits/{tsetmc_code}/{query_date.year}\
{query_date.month:02}{query_date.day:02}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_best_limits_intraday_history_list(
            self,
            tsetmc_code: str,
            query_date: date,
            timeout: int = 3
    ) -> list[BestLimitsHistoryRow]:
        """Get and process instrument historical intraday order book"""
        raw = await self.__get_best_limits_intraday_history_list_raw(
            tsetmc_code=tsetmc_code,
            query_date=query_date,
            timeout=timeout
        )
        return [
            BestLimitsHistoryRow(tsetmc_raw_data=x)
            for x in raw["bestLimitsHistory"]
        ]

    async def __get_index_history_raw(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> dict:
        """Get raw index history"""
        req = await self.__client.get(
            f"api/Index/GetIndexB2History/{tsetmc_code}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_index_history(
            self,
            tsetmc_code: str,
            timeout: int = 3
    ) -> list[IndexDaily]:
        """Get and process index history"""
        raw = await self.__get_index_history_raw(
            tsetmc_code=tsetmc_code,
            timeout=timeout
        )
        return [IndexDaily(tsetmc_raw_data=x) for x in raw["indexB2"]]

    async def __get_instrument_option_info_raw(
            self,
            isin: str,
            timeout: int = 3
    ) -> dict:
        """Get raw instrument option info"""
        req = await self.__client.get(
            f"api/Instrument/GetInstrumentOptionByInstrumentID/{isin}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_instrument_option_info(
            self,
            isin: str,
            timeout: int = 3
    ) -> InstrumentOptionInfo:
        """Get and process instrument option info"""
        raw = await self.__get_instrument_option_info_raw(
            isin=isin,
            timeout=timeout
        )
        return InstrumentOptionInfo(tsetmc_raw_data=raw["instrumentOption"])

    async def __get_primary_market_overview_raw(
            self,
            timeout: int = 3
    ) -> dict:
        """Get raw primary market overview"""
        req = await self.__client.get(
            "api/MarketData/GetMarketOverview/1",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_primary_market_overview(
            self,
            timeout: int = 3
    ) -> PrimaryMarketOverview:
        """Get and process primary market overview"""
        raw = await self.__get_primary_market_overview_raw(
            timeout=timeout
        )
        return PrimaryMarketOverview(
            tsetmc_raw_data=raw["marketOverview"]
        )

    async def __get_secondary_market_overview_raw(
            self,
            timeout: int = 3
    ) -> dict:
        """Get raw secondary market overview"""
        req = await self.__client.get(
            "api/MarketData/GetMarketOverview/2",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_secondary_market_overview(
            self,
            timeout: int = 3
    ) -> SecondaryMarketOverview:
        """Get and process secondary market overview"""
        raw = await self.__get_secondary_market_overview_raw(
            timeout=timeout
        )
        return SecondaryMarketOverview(
            tsetmc_raw_data=raw["marketOverview"]
        )

    async def __get_market_watch_raw(
            self,
            ref_id: int = 0,
            h_even: int = 0,
            timeout: int = 3
    ) -> dict:
        """Get raw market watch page"""
        req = await self.__client.get(
            f"api/ClosingPrice/GetMarketWatch?market=0&paperTypes[0]=1&paperTypes[1]=2\
                &paperTypes[2]=3&paperTypes[3]=4&paperTypes[4]=5&paperTypes[5]=6\
                    &paperTypes[6]=7&paperTypes[7]=8&paperTypes[8]=9&showTraded=false\
                        &withBestLimits=true&hEven={h_even}&RefID={ref_id}",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_market_watch(
            self,
            ref_id: int = 0,
            h_even: int = 0,
            timeout: int = 3
    ) -> list[MarketWatchTradeData]:
        """Get and process market watch page"""
        raw = await self.__get_market_watch_raw(
            ref_id=ref_id,
            h_even=h_even,
            timeout=timeout
        )
        return [
            MarketWatchTradeData(tsetmc_raw_data=x)
            for x in raw["marketwatch"]
        ]

    async def __get_client_type_all_raw(
            self,
            timeout: int = 3
    ) -> dict:
        """Get raw market client type"""
        req = await self.__client.get(
            "api/ClientType/GetClientTypeAll",
            timeout=timeout
        )
        if req.status_code != 200:
            raise TsetmcScrapeException(
                f"Bad response: [{req.status_code}]",
                status_code=req.status_code
            )
        return json.loads(req.text)

    async def get_client_type_all(
            self,
            timeout: int = 3
    ) -> list[MarketWatchClientTypeData]:
        """Get and process market client type"""
        raw = await self.__get_client_type_all_raw(
            timeout=timeout
        )
        return [
            MarketWatchClientTypeData(tsetmc_raw_data=x)
            for x in raw["clientTypeAllDto"]
        ]



"""
Gathers the classes from realtime and other instruments data \
to build complete instrument classes
"""
from dataclasses import dataclass
from abc import ABC
from datetime import date
from tse_utils.models.realtime import (
    OrderBook,
    ClientType,
    TradeCandle,
    DeepOrderBook,
    OrderLimitations,
    BigQuantityParams
)


@dataclass
class InstrumentIdentification:
    """
    Holds the identification for instruments.
    """
    isin: str = None
    tsetmc_code: str = None
    ticker: str = None
    name_persian: str = None
    name_english: str = None
    is_obsolete: bool = False

    def __str__(self):
        return f"{self.ticker} [{self.isin}]"


@dataclass
class InstrumentRealtime(ABC):
    """Interface for Instrument class that contains the realtime data"""
    orderbook: OrderBook = None
    client_type: ClientType = None
    intraday_trade_candle: TradeCandle = None
    deep_orderbook: DeepOrderBook = None

    def __init__(self):
        self.orderbook = OrderBook()
        self.client_type = ClientType()
        self.intraday_trade_candle = TradeCandle()
        self.deep_orderbook = DeepOrderBook()


class Instrument(InstrumentRealtime):
    """
    Holds all available data for a specific tradable instrument.
    """

    def __init__(
        self,
        identification: InstrumentIdentification,
    ):
        self.identification: InstrumentIdentification = identification
        self.big_quantity_params: BigQuantityParams = BigQuantityParams()
        self.order_limitations: OrderLimitations = OrderLimitations()
        InstrumentRealtime.__init__(self=self)

    def ticker_with_tsetmc_hyperlink(self) -> str:
        """
        Returns an HTML element containing a hyperlink to the TSETMC page for instrument.
        """
        return f"<a href=\"http://www.tsetmc.com/Loader.aspx?ParTree=151311\
            &i={self.identification.tsetmc_code}\">{self.identification.ticker}</a>"

    def has_buy_queue(self) -> bool:
        """Checks if instrument has a queue on the buy side"""
        return self.orderbook.rows[0].demand_price == self.order_limitations.max_price

    def has_sell_queue(self) -> bool:
        """Checks if instrument has a queue on the sell side"""
        return self.orderbook.rows[0].supply_price == self.order_limitations.min_price

    def __str__(self):
        return str(self.identification)


class DerivativeInstrument(Instrument):
    '''
    Derivative instrument contains a self.underlying that represents the underlying instrument.
    '''

    def __init__(self, underlying: Instrument, **kwargs):
        self.underlying = underlying
        Instrument.__init__(self=self, **kwargs)


@dataclass
class ExerciseParams:
    """Holds exercise info for derivative instruments that have exercise"""
    exercise_price: int = None
    exercise_date: date = None


@dataclass
class OptionInstrumentMarginParams:
    """Holds parameters used for the calculations of margins"""
    a_factor: float = None
    b_factor: float = None
    c_factor: float = None


class OptionInstrument(
    DerivativeInstrument,
    ExerciseParams,
    OptionInstrumentMarginParams
):
    """Holds data for option contract instruments"""

    def __init__(
            self,
            exercise_date: date,
            exercise_price: int,
            lot_size: int = None,
            **kwargs
    ):
        self.lot_size: int = lot_size
        ExerciseParams.__init__(
            self=self,
            exercise_date=exercise_date,
            exercise_price=exercise_price
        )
        OptionInstrumentMarginParams.__init__(self=self)
        DerivativeInstrument.__init__(self=self, **kwargs)


@dataclass
class IndexIdentification:
    """
    Holds the identification for an index, for example the overal index.
    """
    tsetmc_code: str = None
    persian_name: str = None

    def __str__(self) -> str:
        return f"{self.persian_name} [{self.tsetmc_code}]"


@dataclass
class Index:
    """
    Holds all available data for a specific index.
    """
    identification: IndexIdentification
    min_value: int
    max_value: int
    last_value: int

    def __str__(self) -> str:
        return f"{self.identification}"



"""Building the library"""
import setuptools

setuptools.setup(
    name="tse_utils",
    version="1.1.5",
    author="Arka Equities & Securities",
    author_email="info@arkaequities.com",
    description="General utilities used for data mining in Tehran Stock Exchange (TSE).",
    long_description="Utilities for data mining and auto-trading in Tehran Stock Exhchange. \
        Contains TSETMC API implementation and other useful interfaces for securities market \
        data catching and processing.",
    packages=setuptools.find_packages(),
    install_requires=["httpx", "beautifulsoup4", "lxml"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



"""
This module uses httpx to fetch data asynchronously \
from the TSETMC website. 
"""
import httpx
from bs4 import BeautifulSoup
from tse_utils.tse_client.models import (
    TseClientInstrumentIdentitification,
    TseClientScrapeException
)


class TseClientScraper():
    """
    This class fetches data from tsetmc client.
    """
    base_address: str = "http://service.tsetmc.com/WebService/TseClient.asmx"

    def __init__(self):
        self.__client = httpx.AsyncClient(headers={
            "Host": "service.tsetmc.com",
            "SOAPAction": "http://tsetmc.com/Instrument",
            "accept": "text/xml",
            "Content-type": "text/xml"
        }, base_url=TseClientScraper.base_address)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.__client.aclose()

    async def __get_instruments_list_raw(
            self,
            timeout: int = 3
    ) -> str:
        """Gets instruments list raw"""
        xml_data = """\
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" \
xmlns:xsd="http://www.w3.org/2001/XMLSchema" \
xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
    <soap:Body>
        <Instrument xmlns="http://tsetmc.com/">
            <UserName>string</UserName>
            <Password>string</Password>
            <Flow>unsignedByte</Flow>
        </Instrument>
    </soap:Body>
</soap:Envelope>
"""
        http_content = xml_data.encode(encoding='UTF-8')
        req = await self.__client.post(
            self.base_address,
            content=http_content,
            timeout=timeout
        )
        if req.status_code != 200:
            raise TseClientScrapeException(
                f"Bad response: [{req.status_code}]", status_code=req.status_code)
        return req.text

    async def get_instruments_list(
            self,
            timeout: int = 3
    ) -> tuple[
        list[TseClientInstrumentIdentitification],
        list[TseClientInstrumentIdentitification]
    ]:
        """Get and process instruments list"""
        raw = await self.__get_instruments_list_raw(timeout=timeout)
        element = BeautifulSoup(raw, features="xml").findChild(
            "InstrumentResult").text
        processed = [TseClientInstrumentIdentitification(
            x.split(",")) for x in element.split(";")]
        return [
            x
            for x in processed
            if not x.is_index
        ], [
            x
            for x in processed
            if x.is_index
        ]



"""Test the models in tse_utils library"""
import unittest
from datetime import datetime, date
from tse_utils.models import trader, instrument, realtime, enums


class TestModels(unittest.TestCase):
    """Test the models in tse_utils library"""

    def __init__(self, *args, **kwargs):
        self.sample_instrument = instrument.Instrument(
            instrument.InstrumentIdentification(
                isin="IRO1FOLD0001",
                tsetmc_code="46348559193224090",
                ticker="فولاد"
            ))
        self.sample_date = date(year=2023, month=4, day=30)
        super().__init__(*args, **kwargs)

    def test_initiate_option_instrument(self):
        """Test initiation of an option instrument"""
        sample_option = instrument.OptionInstrument(
            exercise_date=date(year=2023, month=10, day=18),
            exercise_price=1819,
            underlying=self.sample_instrument,
            identification=instrument.InstrumentIdentification(
                isin="IRO9FOLD6831",
                ticker="ضفلا7021",
                tsetmc_code="34929123529054163"
            )
        )
        self.assertTrue(sample_option.exercise_price == 1819)
        self.assertTrue(
            sample_option.underlying.identification.isin == "IRO1FOLD0001"
        )

    def test_portfolio_asset_dynamics(self):
        """Test dynamics of portfolio assets"""
        portfolio = trader.Portfolio()
        portfolio.update_asset(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=100
            ))
        self.assertTrue(portfolio.get_asset_quantity(
            self.sample_instrument.identification.isin) == 100)
        portfolio.update_asset(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=200
            ))
        self.assertTrue(portfolio.get_asset_quantity(
            self.sample_instrument.identification.isin) == 200)
        self.assertTrue(portfolio.get_asset(
            self.sample_instrument.identification.isin).quantity == 200)
        portfolio.remove_asset(isin=self.sample_instrument.identification.isin)
        self.assertTrue(portfolio.get_asset_quantity(
            self.sample_instrument.identification.isin) == 0)
        self.assertTrue(portfolio.get_asset(
            self.sample_instrument.identification.isin) is None)
        portfolio.update_asset(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=300
            ))
        self.assertTrue(portfolio.get_asset_quantity(
            self.sample_instrument.identification.isin) == 300)
        portfolio.empty_asset()
        self.assertTrue(portfolio.get_asset_quantity(
            self.sample_instrument.identification.isin) == 0)
        self.assertTrue(portfolio.get_asset(
            self.sample_instrument.identification.isin) is None)

    def test_portfolio_position_dynamics(self):
        """Test dynamics of portfolio positions"""
        portfolio = trader.Portfolio()
        portfolio.update_position(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=-100
            ))
        self.assertTrue(portfolio.get_position_quantity(
            self.sample_instrument.identification.isin) == -100)
        portfolio.update_position(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=-200
            ))
        self.assertTrue(portfolio.get_position_quantity(
            self.sample_instrument.identification.isin) == -200)
        self.assertTrue(portfolio.get_position(
            self.sample_instrument.identification.isin).quantity == -200)
        portfolio.remove_position(
            isin=self.sample_instrument.identification.isin)
        self.assertTrue(portfolio.get_position_quantity(
            self.sample_instrument.identification.isin) == 0)
        self.assertTrue(portfolio.get_position(
            self.sample_instrument.identification.isin) is None)
        portfolio.update_position(
            trader.PortfolioSecurity(
                isin=self.sample_instrument.identification.isin,
                quantity=-300
            ))
        self.assertTrue(portfolio.get_position_quantity(
            self.sample_instrument.identification.isin) == -300)
        portfolio.empty_position()
        self.assertTrue(portfolio.get_position_quantity(
            self.sample_instrument.identification.isin) == 0)
        self.assertTrue(portfolio.get_position(
            self.sample_instrument.identification.isin) is None)

    def test_deep_order_book_dynamics(self):
        """Test dynamics of deep order books"""
        deep_order_book = realtime.DeepOrderBook()
        # Initiation
        deep_order_book.update_buy_row(2, 200, 950)
        deep_order_book.update_buy_row(3, 300, 900)
        deep_order_book.update_buy_row(1, 100, 1000)
        deep_order_book.update_sell_row(20, 210, 1060)
        deep_order_book.update_sell_row(10, 110, 1010)
        deep_order_book.update_sell_row(30, 310, 1110)
        buy_rows = deep_order_book.get_buy_rows()
        sell_rows = deep_order_book.get_sell_rows()
        self.assertTrue(
            buy_rows[0].price == 1000 and buy_rows[0].volume == 100 and buy_rows[0].num == 1)
        self.assertTrue(
            buy_rows[1].price == 950 and buy_rows[1].volume == 200 and buy_rows[1].num == 2)
        self.assertTrue(
            sell_rows[0].price == 1010 and sell_rows[0].volume == 110 and sell_rows[0].num == 10)
        self.assertTrue(
            sell_rows[1].price == 1060 and sell_rows[1].volume == 210 and sell_rows[1].num == 20)
        # Update
        deep_order_book.remove_buy_row(price=950)
        deep_order_book.remove_buy_row(price=940)
        deep_order_book.remove_sell_row(price=1060)
        deep_order_book.remove_sell_row(price=1070)
        buy_rows = deep_order_book.get_buy_rows()
        sell_rows = deep_order_book.get_sell_rows()
        self.assertTrue(
            buy_rows[1].price == 900 and buy_rows[1].volume == 300 and buy_rows[1].num == 3)
        self.assertTrue(
            sell_rows[1].price == 1110 and sell_rows[1].volume == 310 and sell_rows[1].num == 30)
        # Empty
        deep_order_book.empty_buy_rows()
        deep_order_book.empty_sell_rows()
        buy_rows = deep_order_book.get_buy_rows()
        sell_rows = deep_order_book.get_sell_rows()
        self.assertFalse(buy_rows)
        self.assertFalse(sell_rows)

    def test_trader_order_dynamics(self):
        """Test dynamics of orders"""
        class ImplementedTrader(trader.Trader):
            """A sample from abstract trader class"""

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            async def connect(self) -> None:
                pass

            async def connect_looper(
                    self,
                    interval: int = 3,
                    max_trial=10
            ) -> None:
                pass

            async def disconnect(self) -> None:
                pass

            async def get_server_datetime(self) -> datetime:
                pass

            async def pull_trader_data(self):
                pass

            async def subscribe_instruments_list(
                    self,
                    instruments: list[instrument.Instrument]
            ):
                pass

            async def order_send(self, order: trader.Order):
                pass

            async def order_cancel(self, order: trader.Order):
                pass

            async def order_edit(self, order: trader.Order, quantity: int, price: int):
                pass

        sample_trader = ImplementedTrader(
            credentials=trader.TraderCredentials(
                api=trader.TradingAPI(),
                username="aaa"
            )
        )
        # Initiation
        sample_trader.add_order(trader.Order(
            oms_id=1,
            isin=self.sample_instrument.identification.isin,
            side=enums.TradeSide.BUY,
            quantity=10,
            price=5
        ))
        sample_trader.add_order(trader.Order(
            oms_id=2,
            isin=self.sample_instrument.identification.isin,
            side=enums.TradeSide.SELL,
            quantity=20,
            price=30
        ))
        sample_trader.add_order(trader.Order(
            oms_id=3,
            isin=self.sample_instrument.identification.isin,
            side=enums.TradeSide.BUY,
            quantity=30,
            price=20
        ))
        sample_trader.add_order(trader.Order(
            oms_id=4,
            isin=self.sample_instrument.identification.isin,
            side=enums.TradeSide.BUY,
            quantity=40,
            price=20
        ))
        self.assertTrue(sample_trader.get_order(1).quantity == 10)
        self.assertTrue(sample_trader.get_order(4).quantity == 40)
        self.assertTrue(sample_trader.get_order_custom(
            lambda x: x.side == enums.TradeSide.SELL).oms_id == 2)
        self.assertTrue(sample_trader.get_order_custom(
            lambda x: x.price == 20).oms_id == 3)
        self.assertIsNone(sample_trader.get_order(5))
        self.assertIsNone(sample_trader.get_order_custom(
            lambda x: x.quantity == 100))
        all_orders = sample_trader.get_orders()
        buy_orders = sample_trader.get_orders(
            lambda x: x.side == enums.TradeSide.BUY)
        self.assertTrue(len(all_orders) == 4)
        self.assertTrue(len(buy_orders) == 3)
        # Edition
        sample_order = sample_trader.get_order_custom(
            lambda x: x.quantity == 30)
        sample_order.price = 25
        sample_order.quantity = 35
        self.assertTrue(sample_trader.get_order_custom(
            lambda x: x.price == 20).oms_id == 4)
        self.assertTrue(sample_trader.get_order_custom(
            lambda x: x.quantity == 35).price == 25)
        # Deletion
        sample_trader.remove_order(3)
        self.assertIsNone(sample_trader.get_order_custom(
            lambda x: x.quantity == 35))
        # Emptying
        self.assertTrue(sample_trader.get_orders())
        sample_trader.empty_orders()
        self.assertFalse(sample_trader.get_orders())


if __name__ == '__main__':
    unittest.main()
